 /*******************************************************************
 * CodeRedScanner Version 2 08/07/2001                              *
 * Kirby Kuehl (kkuehl@cisco.com)                                   *
 * Tested on FreeBSD, Linux, and Solaris                            *
 ********************************************************************
 * @(#)coderedscanner.c v2.5 08/16/2001 Kirby Kuehl and Rob Thomas
 * @(#)kkuehl@cisco.com robt@cymru.com
 *
 * 08 AUG 2001 v2.3 Enhancements robt@cymru.com
 * - Added POSIX threads
 *
 * 11 AUG 2001 v2.4 Enhancements kkuehl@cisco.com and robt@cymru.com
 * - Compensated for a threads issue in some Solaris 8 implementations.
 * - Added additional check for IIS 4 servers.
 * - Improved signal handling for the threads.  This included the
 *   addition of pthread_detach() to avoid the lengthy TCP
 *   connection attempt timeout.
 * - Added -c <timeout in seconds> option so that a scan timeout
 *   value could be easily set.  The default timeout is 15 seconds.
 * - Added code to handle 404 error message from GET x.ida.
 * - Testing, testing, testing.  :)
 *
 * 12 AUG 2001 v2.5SUN Enhancements robt@cymru.com
 * - Fixed a bug with name resolution in threads.
 * - This port runs only on Sun due to gethostby*_r()
 *   incompatabilities.  Version 2.5 will include support for Linux,
 *   FreeBSD, and AIX.
 *
 * 16 AUG 2001 v2.5 Enhancements robt@cymru.com
 * - Added -n no name resolution flag and removed name lookups where
 *   appropriate.
 * - Fixed a bug that caused single host scans to take a _very_ long
 *   time.  :)
 * - Moved the single host scans back to threads.
 * - Ported to Linux and AIX.  This involved the joyous task of
 *   figuring out all of the unique gethostby*_r() calls.  Ah,
 *   standards!  :)
 * - Added setrlimit(RLIMIT_NOFILES, &limit) to increase NOFILES,
 *   both current and maximum, to 1024.  This is necessary because
 *   the number of concurrent threads (thus, open descriptors) can
 *   become quite high.
 *
 ********************************************************************/

#define _REENTRANT

#include "coderedscan.h"

typedef struct {
	char ipaddress[1024];
	FILE *output;
} threadWorkOrder_t;

threadWorkOrder_t *workOrder;
int id = 0;
int no_res = 0;
pthread_t peerVersionQuery[255];

void timedout(int signo);
void usage();
int versionquery(threadWorkOrder_t *);

int main(int argc, char *argv[])
{
   extern int optopt;
   extern char *optarg;
   int c = 0;
   int range = 0;
   int list = 0;
   int singlehost =0;
   int max_thread = 1;
   int timeout = 15;
   int high = 0;
   int low = 0;
   int loop = 0;
   int herrno;
   struct sockaddr_in sin; 
   unsigned long start;
   unsigned long end;
   unsigned long counter;
   char outputfile[1024];
   char ipaddress[1024];
   char startip[1024];
   char endip[1024]; 
   char buf[8192];
   struct hostent *host;
   struct hostent result;
   struct in_addr inaddr;
   FILE *input, *output;
   struct rlimit limit;

   limit.rlim_max = 1024;
   limit.rlim_cur = 1024;
   if(setrlimit(RLIMIT_NOFILE, &limit)) {
#ifdef VERBOSE
		fprintf(stderr, "setrlimit() failed for nofiles.\n");
#endif
   }

   if((argc < 3) || (argc > 10))
      usage();
   while((c = getopt(argc,argv,"nh:s:e:l:t:c:")) != -1)
   {
      switch(c)
      {
		 case 'n' : /* No name resolution please */
		 {
			no_res = 1;
			break;
		 }
         case 'h' :  /* Single host scan  */
         {
			if(inet_pton(AF_INET, optarg, &inaddr)) {
				host = NULL;
	            if((start = inet_addr(optarg))==INADDR_NONE) {
					fprintf(stderr,"%s is an invalid IP Address.\n",optarg);
					exit(-1);
				}
			} else {
				if (no_res == 0) {
					host = gethostbyname(optarg);
				} else {
					fprintf(stderr,
					 "ERROR:  Name given but name resolution is disabled!\n\n");
					usage();
					exit(-1);
				}
			}
            sprintf(outputfile, "%s.txt", optarg);
            sprintf(startip, "%s", optarg);
            singlehost = 1;
            break;
         } 
         case 's' :  /* Starting ip address */
         {
            if((start = inet_addr(optarg))==INADDR_NONE)
            {
               fprintf(stderr,"%s is an invalid IP Address.\n",optarg);
               exit(-1);
            }
            sprintf(startip,"%s",optarg);
            break;
         }
         case 'e' : /* Ending ip address */
         {
            if((end = inet_addr(optarg))==INADDR_NONE)
            {
               fprintf(stderr,"%s is an invalid IP Address.\n",optarg);
               exit(-1);
            }
            sprintf(endip,"%s",optarg);
            sprintf(outputfile,"%s-%s.txt",startip,endip);
            range = 1;
            break;
         }
         case 'l' :
         {
            if((input = fopen(optarg, "r")) == NULL)
            {
                printf("Unable to open %s\n",optarg);
                exit(-1);
            }
            sprintf(outputfile,"results-%s",optarg); 
            list = 1;
            break;
         }
		 case 't' : /* Maximum number of concurrent threads */
		 {
			if(atoi(optarg) > 255) {
	 	 	 	fprintf(stderr, "Maximum threads = 255, setting to 255.\n");
				max_thread = 255;
			} else if(atoi(optarg) < 1) {
				max_thread = 1;
	 	 	} else {
				max_thread = atoi(optarg);
			}
	 	 	break;
		 }
		 case 'c' : /* Scan timeout */
		 {
			if(atoi(optarg) < 5) {
			   fprintf(stderr, "Minimum timeout is 5 seconds, setting to 5.\n");
			   timeout = 5;
			} else {
				timeout = atoi(optarg);
			}
			break;
		 }
         default:
         {
            usage(); 
            break;
         }
        
      }
   }

  if((output=fopen(outputfile,"a")) == NULL)
  {
     printf("Unable to open %s for writing.\n",outputfile);
     exit(-1);
  } 

  if(singlehost)
  {
	bzero(&sin, sizeof(sin));
	counter = ntohl(start);

	if (host) {
		memcpy(&sin.sin_addr, host->h_addr_list[0], sizeof(sin.sin_addr));
	} else {
		sin.sin_addr.s_addr = htonl(counter);
	}
	sprintf(ipaddress, "%s", inet_ntoa(sin.sin_addr));
	workOrder = (threadWorkOrder_t *)malloc(sizeof(threadWorkOrder_t));
	sprintf(workOrder->ipaddress, "%s", ipaddress);
	workOrder->output = output;
	pthread_create(&peerVersionQuery[id], NULL, (void *) versionquery,
		(void *) workOrder);
	signal(SIGALRM, timedout);
	alarm(timeout);
	pthread_join(peerVersionQuery[loop], NULL);
  }
 
  if(range)
  {
		 low = (-1 * ntohl(end));
		 high = (-1 * ntohl(start));
		 counter = ntohl(start);
		 /* Fixed bug here. */
		 while (counter < end)
		 {
			if((high-low) < max_thread)
			{
				for(loop = low; loop <= high; loop++)
				{				
					bzero(&sin, sizeof(sin));
					sin.sin_addr.s_addr = htonl(counter);      
					sprintf(ipaddress, "%s", inet_ntoa(sin.sin_addr));
					workOrder =
						(threadWorkOrder_t *)malloc(sizeof(threadWorkOrder_t)); 
					sprintf(workOrder->ipaddress, "%s", ipaddress);
					workOrder->output = output;
					pthread_create(&peerVersionQuery[id], NULL,
						(void *) versionquery, (void *) workOrder);
					id++;
					counter++;
				}

        		signal(SIGALRM, timedout);
        		alarm(timeout);

				for(loop = 0; loop <= id; loop++)
				{				
					pthread_join(peerVersionQuery[loop], NULL);
				}
				low = low + max_thread;
				id = 0;
			} else {
				for(loop = low; loop <= (low+max_thread); loop++)
				{
                    bzero(&sin, sizeof(sin));
                    sin.sin_addr.s_addr = htonl(counter);
                    sprintf(ipaddress,"%s", inet_ntoa(sin.sin_addr));
                    workOrder =
                        (threadWorkOrder_t *)malloc(sizeof(threadWorkOrder_t));
                    sprintf(workOrder->ipaddress, "%s", ipaddress);
                    workOrder->output = output;
                    pthread_create(&peerVersionQuery[id], NULL,
                        (void *) versionquery, (void *) workOrder);
					id++;
					counter++;
				}

        		signal(SIGALRM, timedout);
        		alarm(timeout);

                for(loop = 0; loop <= id; loop++)
                {
					pthread_join(peerVersionQuery[loop], NULL);
				}
				low = low + max_thread;
				id = 0;
			}
		 }
  }
  if(list)
  {
     while(fgets(ipaddress,sizeof(ipaddress),input)!=NULL)
     {
         char *token;
         token = strtok(ipaddress, "\n");
         c = 0; 
         while (token !=NULL)
         {
            switch(c)
            {
                case 0:
                sprintf(ipaddress, "%s",token);
                break;
                default:
                break;
            }
            token =strtok(NULL,"\n");
			
			workOrder =
				(threadWorkOrder_t *)malloc(sizeof(threadWorkOrder_t));
			sprintf(workOrder->ipaddress, "%s", ipaddress);
			workOrder->output = output;
			pthread_create(&peerVersionQuery[id], NULL,
				(void *) versionquery, (void *) workOrder);
            c = c +1;
			id++;
			if(id >= max_thread) {
				for(loop = 0; loop <= id; loop++)
				{
					pthread_join(peerVersionQuery[loop], NULL);
				}
				id = 0;
			}
         }
     }
	 if(id != 0)
	 {
		 for(loop = 0; loop <= id; loop++)
		 {
			pthread_join(peerVersionQuery[loop], NULL);
		 }
	 }
  }

  fprintf(stdout,"Done.\n");
  fprintf(output,"Done.\n");
  fclose(output);
  return(0);
}

int versionquery(threadWorkOrder_t *workOrder)
{
    struct sockaddr_in sin;
    struct hostent *host, result;
    char sendbuff[BUFFSIZE];
    char recvbuff[BUFFSIZE];
	char buf[8192];
    char *line;
    struct in_addr inaddr;
    int sockfd = 0;
    int i  = 0;    
    int iis = 1; 
	int herrno;
    char *uri[] ={
    "HEAD / HTTP/1.0\r\nUser-Agent: HTTPDQuery 1.0\r\n\r\n", 
    "GET /x.ida?AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=X HTTP/1.1\r\nHost: CiscoTest\r\n\r\n",
    "GET /scripts/root.exe?/c+dir\r\n\r\n",
    "GET /msadc/root.exe?/c+dir\r\n\r\n",     
    "GET /c/inetpub/scripts/root.exe?/c+dir\r\n\r\n",
    "GET /d/inetpub/scripts/cmd.exe?/c+dir\r\n\r\n"
    };
    struct timeval tv;

    bzero(&sin,sizeof(sin));

	if (no_res == 0) {
#if defined(__GLIBC__)
	    struct hostent hent;
	    char hbuf[8192];

	    if(inet_pton(PF_INET, workOrder->ipaddress, &inaddr)) {
		    if (gethostbyaddr_r((char *) &inaddr,
			sizeof(inaddr), PF_INET, &hent, hbuf,
			sizeof(hbuf), &host, &herrno) < 0)
		    {
		        host = NULL;
    		    }
		} else {
		    if (gethostbyname_r(workOrder->ipaddress, &hent, hbuf,
				sizeof(hbuf), &host, &herrno) < 0)
		    {
		        host = NULL;
		    }
		}	
#elif defined(sun)
	    if(inet_pton(AF_INET, workOrder->ipaddress, &inaddr)) {
			host = gethostbyaddr_r((char *) &inaddr, sizeof(inaddr), AF_INET,
					&result, buf, sizeof(buf), &herrno);
	    } else {
			host = gethostbyname_r(workOrder->ipaddress, &result, buf,
					sizeof(buf), &herrno);
		}
#elif defined(AIX)
        struct hostent_data *hent_data;
        if(inet_pton(AF_INET, workOrder->ipaddress, &inaddr)) {
            if (gethostbyaddr_r((char *) &inaddr, sizeof(inaddr), AF_INET,
                    host, hent_data) < 0)
            {
                host = NULL;
            }
        } else {
            if (gethostbyname_r(workOrder->ipaddress, host, hent_data) < 0)
            {
                host = NULL;
            }
        }
#else
#error "I don't know the reentrant lookup calls on this system."
#endif
	} else {
		host = NULL;
	}

    if(host) {
		memcpy(&sin.sin_addr, host->h_addr_list[0], sizeof(sin.sin_addr));
    } else {
		if((inet_addr(workOrder->ipaddress))!=INADDR_NONE) {
			sin.sin_addr.s_addr = inet_addr(workOrder->ipaddress);
		} else {
			fprintf(stderr, "Eject!  Eject!  No IP address to scan!\n");
			exit(1);
		}
	}

    for(i=0;i<=5 && (iis==1);i++) 
    {
       if((sockfd = socket(AF_INET,SOCK_STREAM,0)) == -1)
        {
           fprintf(stderr,"Unable to initialize socket.\n");
           return;
        }
        tv.tv_sec  = SNDRCVTIMEOUT;
        tv.tv_usec = SNDRCVTIMEOUT; 
        #if defined (_BSD)
        if(setsockopt(sockfd,SOL_SOCKET,SO_RCVTIMEO,&tv,sizeof(tv))!=0)
           perror("setsockopt");
        if(setsockopt(sockfd,SOL_SOCKET,SO_SNDTIMEO,&tv,sizeof(tv))!=0)
           perror("setsockopt");
        if(setsockopt(sockfd,SOL_SOCKET,SO_RCVLOWAT,(void *) &wait,sizeof(wait))!=0)
           perror("setsockopt");
#endif
        
        sin.sin_family=AF_INET;
        sin.sin_port = htons(80);

        if((connect(sockfd, (struct sockaddr *)&sin, sizeof(sin)) == -1))
        {
           if(errno==EINTR)
              errno=ETIMEDOUT;
			#ifdef VERBOSE
			fprintf(stderr,"%s Closed\n",(char *)inet_ntoa(sin.sin_addr));
			#endif
           close(sockfd);
           return;
        }

		#ifdef VERBOSE
		fprintf(stderr,"%s Open\n",(char *)inet_ntoa(sin.sin_addr));
		#endif
        memset(recvbuff,0,BUFFSIZE);
        memset(sendbuff,0,BUFFSIZE);
        
        sprintf(sendbuff,"%s",uri[i]);
       
        if(send(sockfd,sendbuff,strlen(sendbuff),0) == -1)
        {
           fprintf(stderr, "Error sending request.\n");
           close(sockfd);
           return;
        }
		#ifdef VERBOSE
		else
		fprintf(stderr,"%s",sendbuff);
		#endif
    
        while(recv(sockfd,recvbuff,BUFFSIZE,0)>0)
        {
			#ifdef VERBOSE
			fprintf(stderr,"%s",recvbuff);
			#endif
           
           line = strtok(recvbuff,"\r\n");
           while(line !=NULL)
           {
              if(i == 0) /* Just get Server Version */
              {
                 if(strstr(line,"Server:"))
                 {
                    if(strstr(line,"Microsoft-IIS")) {
						if (no_res == 0) {
							fprintf(stdout,
								"%s [%s]: %s\n**IIS Code Red Checks**\n",
								(char *)inet_ntoa(sin.sin_addr),
								host==NULL ? "unknown" : host->h_name,line);
						} else {
							fprintf(stdout,"%s: %s\n**IIS Code Red Checks***\n",
								(char *)inet_ntoa(sin.sin_addr), line);
						}
                    } else {
						iis = 0;
						if (no_res == 0) {
							fprintf(stdout,"%s [%s]: %s\n**Not Vulnerable**\n",
								(char *)inet_ntoa(sin.sin_addr),
								host==NULL ? "unknown" : host->h_name,line);
						} else {
							fprintf(stdout,"%s: %s\n**Not Vulnerable***\n",
								(char *)inet_ntoa(sin.sin_addr), line);
						}
                    }
                    break;
                 }
              }
              if(i != 0)
              {
                 if(strstr(line,"404 Object Not Found"))
                 {
                    if(i == 1)
                    {
						if (no_res == 0) {
							fprintf(stderr,"%s [%s] is patched: %s\n",
								(char *)inet_ntoa(sin.sin_addr),
								host==NULL ? "unknown" : host->h_name,line);
							fprintf(workOrder->output,"%s [%s] is patched: %s\n"								, (char *)inet_ntoa(sin.sin_addr),
                                host==NULL ? "unknown" : host->h_name,line);
						} else {
							fprintf(stderr,"%s is patched: %s\n",
								(char *)inet_ntoa(sin.sin_addr), line);
							fprintf(workOrder->output,"%s is patched: %s\n",
								(char *)inet_ntoa(sin.sin_addr), line);
						}
                    }
					#ifdef VERBOSE
					fprintf(stderr,"%s %s\n",line,uri[i]);
					#endif 
                    break;
                 } 
                 if(strstr(line,"403 Access Forbidden"))
                 {
					#ifdef VERBOSE
					fprintf(stderr,"%s %s\n",line,uri[i]);
					#endif    
                    break;
                 }

                 if(strstr(line,"0x80040e14"))
                 {
					if (no_res == 0) {
	                    fprintf(stderr,"%s [%s] is patched: %s\n",
							(char *)inet_ntoa(sin.sin_addr),
							host==NULL ? "unknown" : host->h_name,line);
						fprintf(workOrder->output,"%s [%s] is patched: %s\n",
							(char *)inet_ntoa(sin.sin_addr),
							host==NULL ? "unknown" : host->h_name,line);
					} else {
	                    fprintf(stderr,"%s is patched: %s\n",
							(char *)inet_ntoa(sin.sin_addr), line);
						fprintf(workOrder->output,"%s is patched: %s\n",
							(char *)inet_ntoa(sin.sin_addr), line);
					}
                    break;
                 }
                 
                 if(strstr(line,"Error 0xc0000005 caught while processing query"))
                 {
					if (no_res == 0) {
						fprintf(stderr,"%s [%s] is unpatched: %s\n",      
							(char *)inet_ntoa(sin.sin_addr),
							host==NULL ? "unknown" : host->h_name,line);
						fprintf(workOrder->output,"%s [%s] is unpatched: %s\n",
							(char *)inet_ntoa(sin.sin_addr),
							host==NULL ? "unknown" : host->h_name,line);  
					} else {
						fprintf(stderr,"%s is unpatched: %s\n",      
							(char *)inet_ntoa(sin.sin_addr), line);
						fprintf(workOrder->output,"%s is unpatched: %s\n",
							(char *)inet_ntoa(sin.sin_addr), line);
					}
                    break;
                 }
         
                 if(strstr(recvbuff,"Directory of"))
                 {
					if (no_res == 0) {
						fprintf(stderr,"%s [%s] Compromised Variant 2\n",
							(char *)inet_ntoa(sin.sin_addr),
							host==NULL ? "unknown" : host->h_name);
						fprintf(workOrder->output,
							"%s [%s] Compromised Variant 2\n",
							(char *)inet_ntoa(sin.sin_addr),
							host==NULL ? "unknown" : host->h_name);
					} else {
						fprintf(stderr,"%s Compromised Variant 2\n",
							(char *)inet_ntoa(sin.sin_addr), line);
						fprintf(workOrder->output, "%s Compromised Variant 2\n",
							(char *)inet_ntoa(sin.sin_addr), line);
					}
                    break;
                 }
              }
              line = strtok(NULL,"\r\n"); 
           }
        }
         
        close(sockfd);
        fflush(workOrder->output); 
		#ifdef VERBOSE
		fprintf(stderr,"Closing %s\n",(char *)inet_ntoa(sin.sin_addr));
		#endif
    }  /* While Checking uri  */
    return(0);  
}

void timedout(int signo)
{
	int loop = 0;

	signal(SIGALRM, timedout);

	#ifdef VERBOSE
	fprintf(stderr,"Connection timed out.  Detaching %d threads.\n", id);
	#endif

	for(loop = 0; loop <= id; loop++)
	{				
		pthread_detach(peerVersionQuery[loop]);
		#ifdef VERBOSE
		fprintf(stderr, "pthread_detach() sent to thread %d.\n", id);
		#endif
	}
	return;
}

void usage()
{
	fprintf(stderr,"coderescan.c v2.5 08/16/2001 Kirby Kuehl and Rob Thomas\n");
	fprintf(stderr,"Send bugs/suggestions/thanks :) to kkuehl@cisco.com and robt@cymru.com\n\n");
	fprintf(stderr, "Usage: coderedscan [-hseltcn]\n");
	fprintf(stderr, "\t-h <hostname|IP> - Scan a single host.\n");
	fprintf(stderr, "\t-s <start IP> -e <end IP> - Scan a range of hosts.\n");
	fprintf(stderr, "\t-l <file> - Scan a list of IPs contained in <file>.\n");
	fprintf(stderr, "\t-t <threads> - Maximum concurrent threads, no higher than 255.\n");
	fprintf(stderr, "\t-c <seconds> - Timeout in seconds, default 15.\n");
	fprintf(stderr, "\t-n - No name resolution.\n");
	exit(-1);
}

